from rest_framework import serializers
from rest_framework.utils import model_meta
from arkindex.project.serializer_fields import EnumField
from arkindex.dataimport.models import DataImport, DataImportMode, DataImportState, DataFile, Revision
import celery.states


class TaskSerializer(serializers.Serializer):
    """
    Serialize a task and its status info
    """

    id = serializers.UUIDField()
    name = serializers.SerializerMethodField()
    state = serializers.CharField()
    progress = serializers.SerializerMethodField()
    messages = serializers.SerializerMethodField()  # TODO: use a custom serializer ?
    parent_id = serializers.SerializerMethodField()

    def get_name(self, obj):
        name = obj._get_task_meta().get('task_name')

        # Remove prefix
        prefix = 'arkindex.dataimport.tasks.'
        if name and name.startswith(prefix):
            return name[len(prefix):]
        return name

    def get_messages(self, obj):
        return obj._get_task_meta().get('messages', [])

    def get_progress(self, obj):
        if obj.state in celery.states.READY_STATES:
            return 1.0
        if obj.state == 'PROGRESS':
            result = obj._get_task_meta().get('result') or {}
            return result.get('progress', 0.0)
        return 0.0

    def get_parent_id(self, obj):
        return obj.parent and obj.parent.id

    class Meta:
        fields = (
            'id',
            'state',
            'result',
            'parent_id',
        )
        read_only_fields = fields


class DataImportLightSerializer(serializers.ModelSerializer):
    """
    Serialize a data importing workflow
    """

    state = EnumField(DataImportState, read_only=True)
    mode = EnumField(DataImportMode)
    creator = serializers.HiddenField(default=serializers.CurrentUserDefault())

    class Meta:
        model = DataImport
        fields = (
            'id',
            'state',
            'mode',
            'corpus',
            'creator',
        )
        read_only_fields = ('id', )


class ImagesPayloadSerializer(serializers.Serializer):
    """
    Serialize an image importing payload
    """

    folder_name = serializers.CharField()
    volume_name = serializers.CharField()


class DataImportSerializer(DataImportLightSerializer):
    """
    Serialize a data importing workflow with its payload
    """

    # Redefine state as writable
    state = EnumField(DataImportState)
    payload = serializers.JSONField()
    tasks = TaskSerializer(many=True, read_only=True)
    task_count = serializers.IntegerField(read_only=True)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.instance is None:
            return
        payload_serializers = {
            DataImportMode.Images: ImagesPayloadSerializer,
        }
        self.fields['payload'] = payload_serializers.get(self.instance.mode, serializers.JSONField)()

    def update(self, instance, validated_data):
        # ModelSerializer disallows updating with nested serializers.
        # We dynamically inject a payload serializer, but it is useless for the database
        # since the payload field is a JSONField with no particular checks.
        # We don't have to care about a nested serializer and we can treat it just as a dict.
        # This code is copy-pasted from DRF's ModelSerializer.
        info = model_meta.get_field_info(instance)
        for attr, value in validated_data.items():
            if attr in info.relations and info.relations[attr].to_many:
                field = getattr(instance, attr)
                field.set(value)
            else:
                setattr(instance, attr, value)
        instance.save()
        return instance

    class Meta(DataImportLightSerializer.Meta):
        fields = (
            'id',
            'state',
            'mode',
            'corpus',
            'creator',
            'payload',
            'files',
            'tasks',
            'task_count',
            'result',
        )


class DataFileSerializer(serializers.ModelSerializer):
    """
    Serialize a single uploaded file
    """

    class Meta:
        model = DataFile
        fields = (
            'id',
            'name',
            'content_type',
            'size',
        )
        read_only_fields = ('id', 'size', 'content_type', )


class RevisionSerializer(serializers.ModelSerializer):
    """
    Serialize a repository revision
    """

    class Meta:
        model = Revision
        fields = (
            'id',
            'hash',
            'ref',
            'message',
            'author',
            'commit_url',
        )
