from lxml import etree
from arkindex.project.tools import find_closest
from arkindex.documents.models import ElementType, Act, Element, MetaType
from arkindex.dataimport.models import Revision
import logging
import sys
import os
import re
from django.conf import settings


NAMESPACES = {
    'tei': 'http://www.tei-c.org/ns/1.0',
    'xml': 'http://www.w3.org/XML/1998/namespace',
}

logger = logging.getLogger(__name__)


class TeiElement(object):
    optional = ()

    def __init__(self, element):
        assert isinstance(element, etree._Element), 'Not an XML element'
        self.element = element
        self.data = self.parse()

    def __getattr__(self, name):
        if name not in self.data:
            raise AttributeError
        return self.data[name]

    def _find(self, path, many=False):
        # Prepend tei: to all path elements
        path = './tei:' + re.sub(r'(//|/)', r'\1tei:', path)
        method = many and self.element.findall or self.element.find
        return method(path, NAMESPACES)

    def get_text(self, path, dump=False):
        child = self._find(path)
        if child is None:
            return None

        # Dump full content with children
        if dump:
            return etree.tostring(child, pretty_print=True).decode('utf-8')

        # Simply get inner text
        return child.text

    def apply_xslt(self, path, xsltpath):
        child = self._find(path)
        if child is None:
            return None

        transform = etree.XSLT(etree.parse(xsltpath))
        child_tree = etree.ElementTree(child)

        return str(transform(child_tree))

    def get_instance(self, cls, path, many=False):
        child = self._find(path, many)
        if child is None:
            return None
        if many:
            return list(map(cls, child))
        return cls(child)

    @property
    def completion(self):
        '''
        Recursive completion algorithm
        '''
        def avg(items):
            items = list(items)
            if not items:
                return 0
            return sum(items) / len(items)

        def _calc(item):
            name, value = item
            if name in self.optional:
                return 1.0
            if value is None:
                logging.debug('None value {}'.format(name))
                return 0.0
            if isinstance(value, str):
                if not value:
                    logging.debug('Empty string {}'.format(name))
                    return 0.0
                return 1.0
            if isinstance(value, TeiElement):
                return value.completion
            if isinstance(value, list):
                return avg(i.completion for i in value)

            raise Exception('Unsupported type for completion {} : {}'.format(type(item), item))

        return avg(map(_calc, self.data.items()))

    def as_dict(self):
        def _dump(value):
            if isinstance(value, TeiElement):
                return value.as_dict()
            if isinstance(value, list):
                return [v.as_dict() for v in value]
            return value

        return {
            name: _dump(value)
            for name, value in self.data.items()
        }

    def save(self, db_elt, revision):
        assert isinstance(db_elt, Element)
        assert isinstance(revision, Revision)
        # Filter to remove empty data
        parsed_metadatas = {
            name: meta
            for name, meta in self.build_metadata().items()
            if meta[1]
        }

        # Remove deleted metadatas
        db_elt.metadatas.exclude(
            name__in=parsed_metadatas.keys(),
        ).delete()

        # Create/update
        for name, meta in parsed_metadatas.items():
            db_meta, created = db_elt.metadatas.get_or_create(
                name=name,
                defaults={
                    'type': meta[0],
                    'value': meta[1],
                    'revision': revision,
                },
            )
            if created:
                continue
            if (db_meta.type, db_meta.value) == meta:  # Nothing to update
                continue
            db_meta.type, db_meta.value = meta
            db_meta.revision = revision
            db_meta.save()


class Witness(TeiElement):
    optional = ('lang', )

    def parse(self):
        return {
            'repository': self.get_text('msDesc/msIdentifier/repository'),
            'repository_id': self.get_text('msDesc/msIdentifier/idno'),
            'folio': self.get_text('locus[@scheme="folio"]'),
            'id': self.get_text('idno'),
            'lang': self.get_text('lang'),
        }

    def __str__(self):
        out = []
        if self.id:
            out.append(self.id)
        if self.folio:
            out.append('folio {}'.format(self.folio))
        if self.repository and self.repository_id:
            out.append('{} from "{}"'.format(self.repository_id, self.repository))
        return ' - '.join(out)


class Date(TeiElement):
    optional = ('when', )

    def parse(self):
        return {
            'text': self.element.text,
            'when': self.element.attrib.get('when')
        }

    def __str__(self):
        '''
        Convert to a human readable date
        '''
        # TODO: support common formats
        if self.when:
            return self.when
        return self.text


class Term(TeiElement):
    def parse(self):
        return {
            'type': self.element.attrib.get('type'),
            'text': self.element.text,
        }


class Text(TeiElement):

    def __str__(self):
        return str(self.witness) or 'Unknown text'

    def parse(self):
        return {
            'id': self.element.attrib.get('{{{}}}id'.format(NAMESPACES['xml'])),
            'witness': self.get_instance(Witness, 'front//witness'),
            'date': self.get_instance(Date, 'front/docDate/date'),
            'location': self.get_text('front/docDate/origPlace'),
            'argument': self.apply_xslt('front/argument', settings.TEI_XSLT_PATH),
            'terms': self.get_instance(Term, 'front/index/term', many=True),
        }

    def build_metadata(self):
        '''
        Build metadatas for act in database
        '''
        def _clean_terms(term_type):
            return '\n'.join(filter(None, [
                re.sub(r'\n\s*', ' ', t.text).strip()
                for t in self.terms
                if t.type == term_type
            ]))

        return {
            'id': (MetaType.Text, self.id),
            'folio': (MetaType.Text, self.witness and self.witness.folio),
            'name': (MetaType.Text, self.witness and self.witness.id),
            'language': (MetaType.Text, self.witness and self.witness.lang),
            'date': (MetaType.Date, self.date and str(self.date)),
            'location': (MetaType.Location, self.location),
            'summary': (MetaType.HTML, self.argument),

            # Future entities
            'persons': (MetaType.Text, _clean_terms('person')),
            'places': (MetaType.Text, _clean_terms('place')),
            'subjects': (MetaType.Text, _clean_terms('subject')),
        }


class Tei(TeiElement):
    def parse(self):
        return {
            'witness': self.get_instance(Witness, 'teiHeader//witness'),
            'texts': self.get_instance(Text, 'text/group/text', many=True),
        }

    def __str__(self):
        return 'TEI {}'.format(self.witness or 'unknown')

    def build_metadata(self):
        '''
        Build metadatas for Volume in database
        '''
        return {
            'name': (MetaType.Text, self.witness.repository_id),
            'repository': (MetaType.Text, self.witness.repository)
        }


class Corpus(TeiElement):
    def parse(self):
        return {
            'tei': self.get_instance(Tei, 'TEI', many=True),
            'title': self.get_text('teiHeader//title[@type="eadid"]'),
        }

    def __str__(self):
        return 'Corpus {}'.format(self.title or 'unknown')


class TeiParser(object):
    def __init__(self, path):
        assert os.path.exists(path), \
            'Invalid path {}'.format(path)

        root = etree.parse(path).getroot()
        # Some files only have xmlns="..." and not xmlns:tei="...", the former maps to None
        assert root.nsmap.get('tei') == NAMESPACES['tei'] or root.nsmap.get(None) == NAMESPACES['tei'], \
            'Missing tei XML namespace'

        # Start parsing
        self.corpus = Corpus(root)

    def check(self):
        logger.info(self.corpus)
        for tei in self.corpus.tei:
            logger.info('{} - completion {:.1%}'.format(tei, tei.completion))

    def match_database(self, corpus):
        '''
        Link TEI elements with DB elements
        Outputs a list of tuples (db_element, tei_element)
        '''
        volumes = corpus.elements.filter(type=ElementType.Volume)
        assert volumes.count() > 0, 'No volumes available'

        # Match volumes
        out = []
        for tei in self.corpus.tei:
            if not tei.witness:
                logger.warning('No witness in {}'.format(str(tei)))
                continue

            tei_name = tei.witness.id or tei.witness.repository_id
            volume = find_closest(tei_name, volumes)
            if not volume:
                logger.warning('No match for {}'.format(tei))
                continue

            out.append((volume, tei))
            logger.info('Matched {} with {}'.format(volume, tei))

            # Load volume acts
            volume_acts = Element.objects.get_descending(volume.id, type=ElementType.Act)
            if not volume_acts.exists():
                logger.warning('No acts in DB for {}'.format(volume))
                continue

            # Match acts
            for text in tei.texts:
                if text.witness is None:
                    logger.warning('No witness on text, skipping.')
                    continue

                act = Act.objects.filter(
                    id__in=volume_acts,
                    number=text.witness.id,
                ).first()
                if act:
                    out.append((act, text))
                    logger.info('Matched {} with {}'.format(act, text))
                else:
                    logger.warning('No match for {}'.format(text))

        return out


if __name__ == '__main__':
    t = TeiParser(sys.argv[1])
    t.check()
